#include <ros/ros.h>

#include "cola2_lib/cola2_rosutils/DiagnosticHelper.h"
#include <cola2_lib/cola2_io/SerialPort.h>
#include <cola2_control/Setpoints.h>
#include <cola2_control/ThrustersInfo.h>

#include <std_srvs/Empty.h>

#include <iostream>
#include <stdint.h>
#include <string>
#include <signal.h>
#include <cmath>
#include <sstream>
#include <iomanip>
#include <boost/bind.hpp>
#include <boost/date_time.hpp>
#include <boost/thread.hpp>

// Serial port
cola2::io::SerialPort serial_port ;

class FINS_485
{
private:
    // Var. used to store the name of the node
    std::string _node_name ;

    // Serial port mutex
    boost::mutex serial_mutex ;

public:
    FINS_485( std::string node_name )
    {
        _node_name = node_name ;
    } ;

    ~FINS_485()
    {
    } ;

    void disableFins(unsigned char address)
    {
        std::vector<unsigned char> fins_command ;
        fins_command.push_back(0xfe) ;
        fins_command.push_back(address) ;
        fins_command.push_back(0x00) ;
        fins_command.push_back(0x00) ;

        // Send
        try
        {
            // Mutex
            boost::mutex::scoped_lock serial_lock(serial_mutex) ;

            serial_port.write(fins_command, _node_name, 100.0) ;
        }
        catch (std::exception& e)
        {
            std::cerr << _node_name << ": unable to send command: " << e.what() << std::endl ;
        }
    }

    void enableFins(unsigned char address)
    {
        std::vector<unsigned char> fins_command ;
        fins_command.push_back(0xfe) ;
        fins_command.push_back(address) ;
        fins_command.push_back(0x01) ;
        fins_command.push_back(0x01) ;

        // Send
        try
        {
            // Mutex
            boost::mutex::scoped_lock serial_lock(serial_mutex) ;

            serial_port.write(fins_command, _node_name, 100.0) ;
        }
        catch (std::exception& e)
        {
            std::cerr << _node_name << ": unable to send command: " << e.what() << std::endl ;
        }
    }

    bool computeFinsSetpoint(unsigned char address, unsigned char left, unsigned char right)
    {
        std::vector<unsigned char> fins_command ;
        fins_command.push_back(0xfd) ;
        fins_command.push_back(address) ;
        fins_command.push_back(left) ;
        fins_command.push_back(right) ;
        fins_command.push_back(address xor left xor right) ;

        // Send
        try
        {
            // Mutex
            boost::mutex::scoped_lock serial_lock(serial_mutex) ;

            serial_port.write(fins_command, _node_name, 100.0) ;
        }
        catch (std::exception& e)
        {
            std::cerr << _node_name << ": unable to send command: " << e.what() << std::endl ;
        }
    }

    //
    bool waterRequest(unsigned char address)
    {
        std::vector<unsigned char> fins_command ;
        fins_command.push_back(0xfe) ;
        fins_command.push_back(address) ;
        fins_command.push_back(0x02) ;
        fins_command.push_back(0x02) ;

        bool answer = false ;

        try
        {
            // Mutex
            boost::mutex::scoped_lock serial_lock(serial_mutex) ;

            serial_port.write(fins_command, _node_name, 100.0) ;

            std::vector<unsigned char> line ;

            // Compute request
            line.resize(0) ;
            unsigned char nextChar = 0 ;
            int iter ;
            for (iter = 0 ; iter < 4 ; iter++)
            {
                nextChar = serial_port.readByte(100) ;
                line.push_back(nextChar) ;
            }

            // Check protocol
            if (line.size() == 4 and line[1] == address and line[2] == line[3] and line[2] == 1)
            {
                answer = true ;
            }
        }
        catch (std::exception& e)
        {
            std::cerr << _node_name << ": unable to ask for water: " << e.what() << std::endl ;
        }
        return answer ;
    }
} ;

class THRUSTERS_485
{
private:
    // Var. used to store the name of the node
    std::string _node_name ;

    // Serial port mutex
    boost::mutex serial_mutex ;

    // Var. used to store old setpoints
    std::vector<double> old_setpoints ;

    // Var. used to alternate requests
    unsigned int request ;

    // Var. used to alternate thrusters
    unsigned int alternate ;

public:

    THRUSTERS_485( std::string node_name )
    {
        _node_name = node_name ;

        request = 0 ;

        alternate = 0 ;
    } ;
    ~THRUSTERS_485()
    {
    } ;

    struct ThrusterInfo
    {
        std::vector<double> thruster_voltages ;
        std::vector<double> thruster_currents ;
        std::vector<double> thruster_rpms ;
        std::vector<double> thruster_controller_temperatures ;
        std::vector<double> thruster_output_powers ;
        std::vector<std::string> thruster_warnings ;
        std::vector<std::string> thruster_errors ;
        double total_ms_in_driver ;
        std::vector<double> setpoints ;
        std::vector<double> processed_setpoints ;
    } ;
    ThrusterInfo thrusters_info;

    void init(std::vector<unsigned char> thrusters_addresses)
    {
        // Sending something to init thrusters
        for (int i = 0 ; i < thrusters_addresses.size() ; i++)
        {
            computeSetpoint(thrusters_addresses[i], 0.01) ;
        }

        // Vars used to check thrusters
        std::vector<int> zero_setpoint_counter ;
        std::vector<int> bad_current_counter ;
        int i ;
        for (i = 0 ; i < thrusters_addresses.size() ; i++)
        {
            zero_setpoint_counter.push_back(0) ;
            bad_current_counter.push_back(0) ;
            old_setpoints.push_back(0.0) ;

            thrusters_info.thruster_voltages.push_back(0.0) ;
            thrusters_info.thruster_currents.push_back(0.0) ;
            thrusters_info.thruster_rpms.push_back(0.0) ;
            thrusters_info.thruster_controller_temperatures.push_back(0.0) ;
            thrusters_info.thruster_output_powers.push_back(0.0) ;
            thrusters_info.thruster_warnings.push_back("Not received yet") ;
            thrusters_info.thruster_errors.push_back("Not received yet") ;
        }
    }

    /**
     * Send null setpoint method
     *
     * @author Eduard
     * @date 25/2/2013
     */
    void sendZeros(std::vector<unsigned char> thrusters_addresses)
    {
        std::cout << _node_name << ": sending zero" << std::endl ;

        for (int i = 0 ; i < thrusters_addresses.size() ; i++)
        {
            computeSetpoint(thrusters_addresses[i], 0.0) ;
        }
    }

    /**
     * Compute thruster setpoint method
     *
     * @author Eduard
     * @date 25/2/2013
     */
    bool computeSetpoint(unsigned char address, double double_setpoint)
    {
        try
        {
            std::vector<unsigned char> thruster_command ;
            unsigned int int_setpoint ;
            uint16_t converted_setpoint ;
            uint8_t low ;
            uint8_t high ;

            if (double_setpoint < -1.0)
            {
                double_setpoint = -1.0 ;
            }

            if (double_setpoint > 1.0)
            {
                double_setpoint = 1.0 ;
            }

            int_setpoint = round(1024 * double_setpoint * 0.65) ;

            if (int_setpoint < 0)
            {
                converted_setpoint = ~((uint16_t)(-int_setpoint)) + 0b1 ;
            }
            else
            {
                converted_setpoint = (uint16_t)int_setpoint ;
            }

            low = (uint8_t)((converted_setpoint & 0b0000000011111111) >> 0) ;
            high = (uint8_t)((converted_setpoint & 0b1111111100000000) >> 8) ;

            thruster_command.push_back(0xfd) ;
            thruster_command.push_back(address) ;
            thruster_command.push_back(low) ;
            thruster_command.push_back(high) ;
            thruster_command.push_back(address xor low xor high) ;

            // Scope for the mutex
            {
                // Mutex
                boost::mutex::scoped_lock serial_lock(serial_mutex) ;

                // Send
                try
                {
                    serial_port.write(thruster_command, _node_name, 100.0) ;
                }
                catch (std::exception& e)
                {
                    std::cerr << _node_name << ": unable to send command: " << e.what() << std::endl ;
                }

                // Sleep some time
                boost::this_thread::sleep(boost::posix_time::microseconds(1000));

                // Empty serial port buffer, since some drivers return ff byte after receiving command
                bool empty = false ;
                int number_of_bytes_after_setpoint = 0 ;

                while (!empty)
                {
                    try
                    {
                        serial_port.readByte(1) ;
                        number_of_bytes_after_setpoint += 1 ;
                    }
                    catch (std::exception&)
                    {
                        empty = true ;
                    }
                }
            }
            return true ;
        }
        catch (std::exception& e)
        {
            std::cout << _node_name << ": exception while computing setpoint: " << e.what() << std::endl ;
            return false ;
        }
    }

    /**
     * Read info request from serial port
     *
     * @author Eduard
     * @date 25/2/2013
     */
    bool serialComputeRequest(std::vector<unsigned char>& line, unsigned char address, unsigned char request)
    {
        try
        {
            boost::mutex::scoped_lock serial_lock(serial_mutex) ;

            // Empty serial port buffer
            bool empty = false ;
            while (!empty)
            {
                try
                {
                    serial_port.readByte(1) ;
                }
                catch (std::exception&)
                {
                    empty = true ;
                }
            }

            // Send request
            std::vector<unsigned char> thruster_command ;
            thruster_command.push_back(0xfe) ;
            thruster_command.push_back(address) ;
            thruster_command.push_back(request) ;
            thruster_command.push_back(request) ;

            // Send
            try
            {
                serial_port.write(thruster_command, _node_name, 100.0) ;
            }
            catch (std::exception& e)
            {
                std::cerr << _node_name << ": unable to send command: " << e.what() << std::endl ;
            }

            // Compute request
            bool line_ok = false ;
            line.resize(0) ;
            unsigned char nextChar = 0 ;
            int iter ;
            for (iter = 0 ; iter < 6 ; iter++)
            {
                nextChar = serial_port.readByte(100) ;
                line.push_back(nextChar) ;
            }

            if (line.size() == 6)
            {
                // Check checksum, request and address
                if ((line[5] == (line[1] xor line[2] xor line[3] xor line[4])) && (line[2] == request) && (line[1] == address))
                {
                    line_ok = true ;
                }
            }

            return line_ok ;
        }
        catch (std::exception& e)
        {
            std::cout << _node_name << ": exception while computing request: " << e.what() << std::endl ;
            return false ;
        }
    }


    /**
     * Thruster voltage request
     *
     * @author Eduard
     * @date 25/2/2013
     */
    bool computeVoltageRequest(unsigned char address, double& parameter)
    {
        try
        {
            std::vector<unsigned char> line ;

            if (serialComputeRequest(line, address, 0x01))
            {
                parameter = 0.1 * (double)(line[3] + 256 * line[4]) ;
                return true ;
            }
            else
            {
                return false ;
            }
        }
        catch (std::exception&)
        {
            return false ;
        }
    }


    /**
     * Thruster current request
     *
     * @author Eduard
     * @date 25/2/2013
     */
    bool computeCurrentRequest(unsigned char address, double& parameter)
    {
        try
        {
            std::vector<unsigned char> line ;

            if (serialComputeRequest(line, address, 0x02))
            {
                parameter = 0.1 * (double)checkTwosComplement(line[3], line[4]) ;
                return true ;
            }
            else
            {
                return false ;
            }
        }
        catch (std::exception&)
        {
            return false ;
        }
    }


    /**
     * Thruster rpm request
     *
     * @author Eduard
     * @date 25/2/2013
     */
    bool computeRpmRequest(unsigned char address, double& parameter)
    {
        try
        {
            std::vector<unsigned char> line ;

            if (serialComputeRequest(line, address, 0x03))
            {
                parameter = 10.0 * (double)(line[3] + 256 * line[4]) ;
                return true ;
            }
            else
            {
                return false ;
            }
        }
        catch (std::exception&)
        {
            return false ;
        }
    }


    /**
     * Controller temperature request
     *
     * @author Eduard
     * @date 25/2/2013
     */
    bool computeTemperatureRequest(unsigned char address, double& parameter)
    {
        try {
            std::vector<unsigned char> line ;

            if (serialComputeRequest(line, address, 0x04))
            {
                parameter = (double)checkTwosComplement(line[3], line[4]) ;
                return true ;
            }
            else
            {
                return false ;
            }
        }
        catch (std::exception&)
        {
            return false ;
        }
    }


    /**
     * Thruster output power request
     *
     * @author Eduard
     * @date 25/2/2013
     */
    bool computeOutputPowerRequest(unsigned char address, double& parameter)
    {
        try
        {
            std::vector<unsigned char> line ;

            if (serialComputeRequest(line, address, 0x09))
            {
                parameter = (double)checkTwosComplement(line[3], line[4]) ;
                return true ;
            }
            else
            {
                return false ;
            }
        }
        catch (std::exception&)
        {
            return false ;
        }
    }


    /**
     * Thruster warning request
     *
     * @author Eduard
     * @date 25/2/2013
     */
    bool computeWarningRequest(unsigned char address, std::string& parameter)
    {
        try
        {
            std::vector<unsigned char> line ;
            bool valid = false ;

            if (serialComputeRequest(line, address, 0x0a))
            {
                std::stringstream message ;

                if (line[3] == 0)
                {
                    message<<"All OK. " ;
                    valid = true ;
                }
                if ((line[3] & 0x01) == 0x01)
                {
                    message<<"Low voltage. " ;
                    std::cerr << _node_name << ": warning request -> Low voltage" << std::endl ;
                    valid = true ;
                }
                if ((line[3] & 0x02) == 0x02)
                {
                    message<<"High current. " ;
                    std::cerr << _node_name << ": warning request -> High current" << std::endl ;
                    valid = true ;
                }
                if ((line[3] & 0x04) == 0x04)
                {
                    message<<"High controller temperature. " ;
                    std::cerr << _node_name << ": warning request -> High controller temperature" << std::endl ;
                    valid = true ;
                }
                if ((line[3] & 0x08) == 0x08)
                {
                    message<<"High battery temperature. " ;
                    std::cerr << _node_name << ": warning request -> High battery temperature" << std::endl ;
                    valid = true ;
                }
                if ((line[3] & 0x10) == 0x10)
                {
                    message<<"High motor temperature. " ;
                    std::cerr << _node_name << ": warning request -> High motor temperature" << std::endl ;
                    valid = true ;
                }

                if (!valid)
                {
                    message<<"??? " ;
                    std::cerr << _node_name << ": warning request returns invalid response" << std::endl ;
                }

                parameter = message.str().erase(message.str().size() - 1) ;
                return true ;
            }
            else
            {
                return false ;
            }
        }
        catch (std::exception&)
        {
            return false ;
        }
    }


    /**
     * Thruster error request
     *
     * @author Eduard
     * @date 25/2/2013
     */
    bool computeErrorRequest(unsigned char address, std::string& parameter)
    {
        try
        {
            std::vector<unsigned char> line ;
            bool valid = false ;

            if (serialComputeRequest(line, address, 0x0b))
            {
                std::stringstream message ;

                if (line[3] == 0)
                {
                    message<<"All OK. " ;
                    valid = true ;
                }
                if ((line[3] & 0x01) == 0x01)
                {
                    message<<"Signal lost. " ;
                    std::cerr << _node_name << ": error request -> Signal lost" << std::endl ;
                    valid = true ;
                }
                if ((line[3] & 0x02) == 0x02)
                {
                    message<<"Waiting for neutral. " ;
                    std::cerr << _node_name << ": error request -> Waiting for neutral" << std::endl ;
                    valid = true ;
                }
                if ((line[3] & 0x04) == 0x04)
                {
                    message<<"Settings changed (PC). " ;
                    std::cerr << _node_name << ": error request -> Settings changed (PC)" << std::endl ;
                    valid = true ;
                }
                if ((line[3] & 0x08) == 0x08)
                {
                    message<<"Memory failed. " ;
                    std::cerr << _node_name << ": error request -> Memory failed" << std::endl ;
                    valid = true ;
                }
                if ((line[3] & 0x10) == 0x10)
                {
                    message<<"Settings changed (ESC). " ;
                    std::cerr << _node_name << ": error request -> Settings changed (ESC)" << std::endl ;
                    valid = true ;
                }
                if ((line[3] & 0x20) == 0x20)
                {
                    message<<"Motor hall sensor error. " ;
                    std::cerr << _node_name << ": error request -> Motor hall sensor error" << std::endl ;
                    valid = true ;
                }
                if ((line[3] & 0x40) == 0x40)
                {
                    message<<"Internal voltage error. " ;
                    std::cerr << _node_name << ": error request -> Internal voltage error" << std::endl ;
                    valid = true ;
                }
                if ((line[3] & 0x80) == 0x80)
                {
                    message<<"Motor hall learning OK. " ;
                    std::cerr << _node_name << ": error request -> Motor hall learning OK" << std::endl ;
                    valid = true ;
                }
                if ((line[4] & 0x01) == 0x01)
                {
                    message<<"Battery temperature sensor. " ;
                    std::cerr << _node_name << ": error request -> Battery temperature sensor" << std::endl ;
                    valid = true ;
                }
                if ((line[4] & 0x02) == 0x02)
                {
                    message<<"Motor temperature sensor. " ;
                    std::cerr << _node_name << ": error request -> Motor temperature sensor" << std::endl ;
                    valid = true ;
                }

                if (!valid)
                {
                    message<<"??? " ;
                    std::cerr << _node_name << ": error request returns invalid response" << std::endl ;
                }

                // Remove last space
                parameter = message.str().erase(message.str().size() - 1) ;
                return true ;
            }
            else
            {
                return false ;
            }
        }
        catch (std::exception&)
        {
            return false ;
        }
    }

    /**
     * Two's complement if needed
     *
     * @author Eduard
     * @date 25/2/2013
     */
    int checkTwosComplement(const unsigned char& low, const unsigned char& high)
    {
        uint16_t both = 0 ;
        if ((high>>7) == 1)
        {
            both = both | (uint16_t)high ;
            both = both<<8 ;
            both = both | (uint16_t)low ;
            return -(int)(both xor 0xffff) - 1 ;
        }
        return low + high * 256 ;
    }

    void thrusterSetpointCallback(std::vector<double> thrusters_setpoints, std::vector<unsigned char> thrusters_addresses, std::vector<double> thrusters_signs, std::vector<double> thrusters_limiter, bool thrusters_symmetric, double thrusters_max_step, bool thrusters_request_info)
    {
        // Get the initial time
        boost::posix_time::ptime previousTime = boost::posix_time::microsec_clock::universal_time() ;

        // Compute thrusters
        if (thrusters_addresses.size() == thrusters_setpoints.size())
        {
            std::vector<double> setpoints ;
            setpoints.resize(thrusters_addresses.size()) ;

            // Compute each thruster
            int i ;
            for (i = 0 ; i < thrusters_addresses.size() ; i++)
            {
                // Compute signs
                setpoints[i] = (double)thrusters_setpoints[i] * thrusters_signs[i] ;

                // Limiter
                if (setpoints[i] > thrusters_limiter[i])
                {
                    setpoints[i] = thrusters_limiter[i] ;
                    std::cout << _node_name << ": limiting thruster " << i << std::endl ;
                }
                else if (setpoints[i] < -thrusters_limiter[i])
                {
                    setpoints[i] = -thrusters_limiter[i] ;
                    std::cout << _node_name << ": limiting thruster " << i << std::endl ;
                }

                // Derivative filter
                if (thrusters_symmetric)
                {
                    // Filter in both directions
                    if (setpoints[i] > old_setpoints[i] + thrusters_max_step)
                    {
                        setpoints[i] = old_setpoints[i] + thrusters_max_step ;
                    }
                    else if (setpoints[i] < old_setpoints[i] - thrusters_max_step)
                    {
                        setpoints[i] = old_setpoints[i] - thrusters_max_step ;
                    }
                }
                else
                {
                    // Non symmetric. Filter only when increasing setpoint
                    if ((setpoints[i] > old_setpoints[i] + thrusters_max_step) && (setpoints[i] > thrusters_max_step))
                    {
                        if ((old_setpoints[i] + thrusters_max_step) > thrusters_max_step)
                        {
                            setpoints[i] = old_setpoints[i] + thrusters_max_step ;
                        }
                        else
                        {
                            setpoints[i] = thrusters_max_step ;
                        }
                    }
                    else if ((setpoints[i] < old_setpoints[i] - thrusters_max_step) && (setpoints[i] < -thrusters_max_step))
                    {
                        if ((old_setpoints[i] - thrusters_max_step) < -thrusters_max_step)
                        {
                            setpoints[i] = old_setpoints[i] - thrusters_max_step ;
                        }
                        else
                        {
                            setpoints[i] = -thrusters_max_step ;
                        }
                    }
                }

                // Store old setpoints
                old_setpoints[i] = setpoints[i] ;

                // Send setpoints
                if (!computeSetpoint(thrusters_addresses[i], setpoints[i]))
                {
                    std::cerr << _node_name << ": unable to send setpoint, thruster " << i << std::endl ;
                }

                // Request info
                if (thrusters_request_info)
                {
                    // Current request at 10 Hz
                    if (!computeCurrentRequest(thrusters_addresses[i], thrusters_info.thruster_currents[i]))
                    {
                        std::cout << _node_name << ": unable to compute current request, thruster " << i << std::endl ;
                    }

                    // RPM request at 10 Hz
                    if (!computeRpmRequest(thrusters_addresses[i], thrusters_info.thruster_rpms[i]))
                    {
                        std::cout << _node_name << ": unable to compute rpm request, thruster " << i << std::endl ;
                    }

                    // Request each 1.5 seconds
                    if (alternate == i)
                    {
                        switch (request)
                        {
                            case 0:
                                if (!computeVoltageRequest(thrusters_addresses[i], thrusters_info.thruster_voltages[i]))
                                {
                                    std::cout << _node_name << ": unable to compute voltage request, thruster " << i << std::endl ;
                                }
                            break;

                            case 1:
                                // Save some time here...
                                thrusters_info.thruster_output_powers[i] = std::abs(thrusters_info.thruster_voltages[i] * thrusters_info.thruster_currents[i]) ;
                                //if (!computeOutputPowerRequest(thrusters_addresses[i], thrusters_info.thruster_output_powers[i]))
                                //{
                                //    std::cout << _node_name << ": unable to compute output power request, thruster " << i << std::endl ;
                                //}
                            break;

                            case 2:
                                if (!computeWarningRequest(thrusters_addresses[i], thrusters_info.thruster_warnings[i]))
                                {
                                    std::cout << _node_name << ": unable to compute warning request, thruster " << i << std::endl ;
                                }
                            break;

                            case 3:
                                if (!computeErrorRequest(thrusters_addresses[i], thrusters_info.thruster_errors[i]))
                                {
                                    std::cout << _node_name << ": unable to compute error request, thruster " << i << std::endl ;
                                }
                            break;

                            case 4:
                                if (!computeTemperatureRequest(thrusters_addresses[i], thrusters_info.thruster_controller_temperatures[i]))
                                {
                                    std::cout << _node_name << ": unable to compute controller temperature request, controller " << i << std::endl ;
                                }
                            break;

                            default:
                                request = 4 ;
                            break;
                        }

                        if (request != 4 )
                        {
                            request += 1 ;
                        }
                        else
                        {
                            request = 0 ;

                            // Change alternate var.
                            alternate += 1 ;
                            if (alternate == thrusters_addresses.size())
                            {
                                alternate = 0 ;
                            }
                        }
                    }
                }
            }

            // Compute time
            double dt = 0.001 * boost::posix_time::time_period(
                        previousTime, boost::posix_time::microsec_clock::universal_time()).length().total_microseconds() ;
            if (dt > 100.0)
            {
                std::cout << _node_name << ": slow serial port waiting for requests. Total elapsed time: " << dt << "ms" << std::endl ;
            }

            thrusters_info.total_ms_in_driver = dt ;
            thrusters_info.setpoints = thrusters_setpoints ;
            thrusters_info.processed_setpoints = setpoints ;
        }
        else
        {
            std::cerr << _node_name << ": invalid setpoints lenght" << std::endl ;
        }
    }
} ;

class ACTUATORS_ROS
{
private:
    // Node handle
    ros::NodeHandle n ;

    // Timer
    ros::Timer timer ;
    ros::Timer timer_water ;

    // Diagnostics
    cola2::rosutils::DiagnosticHelper diagnostic ;

    // Fins subscriber
    ros::Subscriber fins_sub ;

    // Thruster subscriber
    ros::Subscriber thruster_sub ;

    // Thruster publisher
    ros::Publisher thruster_info_pub ;

    // Message type to publish
    cola2_control::ThrustersInfo thrusters_info ;

    // Pointer to FINS_485 class
    FINS_485 * _fins_ptr ;

    // Pointer to THRUSTERS_485 class
    THRUSTERS_485 * _thrusters_ptr ;

    // Var. used to store the name of the node
    std::string _node_name ;

    // Var. used to store times
    boost::posix_time::ptime lastSetpointCallback ;

    // Var. used to store the state of the fins
    int _fins_enabled ;

    // Var. used to store water detection stat
    bool _water ;

    // Var. used to avoid false water detections
    int _water_offset ;

    // Var. used to store fins_address
    unsigned char _fins_address ;

    // Var. used to store old fins setpoints
    std::vector<double> _old_fins_setpoints ;

    // Device configuration
    struct DevConfig
    {
        std::string sp_path ;
        int sp_baud_rate ;
        int sp_char_size ;
        int sp_stop_bits ;
        std::string sp_parity ;
        std::string sp_flow_control ;

        bool request_info ;
        double max_step ;
        bool symmetric ;
        std::vector<unsigned char> addresses ;
        std::vector<double> signs ;
        std::vector<double> limiter ;
    } ;
    DevConfig device_config ;

public:

    ACTUATORS_ROS( std::string node_name, FINS_485 *fins_485_ptr, THRUSTERS_485 *thrusters_485_ptr ):
        diagnostic(n, node_name, "none"),
        _water(false),
        _water_offset(0),
        _old_fins_setpoints(2)
    {
        _node_name = node_name ;

        _fins_ptr = fins_485_ptr ;
        _thrusters_ptr = thrusters_485_ptr ;
    } ;

    ~ACTUATORS_ROS()
    {
        delete _fins_ptr ;
        delete _thrusters_ptr ;
    } ;

    void init()
    {
        ACTUATORS_ROS::getConfig() ;
        ACTUATORS_ROS::openSerialPort() ;

        // Enable fins with actuators driver start up
        _fins_ptr->enableFins(_fins_address) ;

        // Config to thrusters
        _thrusters_ptr->init(device_config.addresses) ;

        // Fins subscriber
        fins_sub = n.subscribe("/cola2_control/fins_data", 3, &ACTUATORS_ROS::finsSetpointCallback, this) ;

        // Thruster subscriber
        thruster_sub = n.subscribe("/cola2_control/thrusters_data", 3, &ACTUATORS_ROS::thrustersSetpointCallback, this) ;

        // Thruster publisher
        thruster_info_pub = n.advertise<cola2_control::ThrustersInfo>("/cola2_control/thrusters_info" , 3) ;

        // Timer
        lastSetpointCallback = boost::posix_time::microsec_clock::universal_time() ;
        timer = n.createTimer(ros::Duration(0.1), &ACTUATORS_ROS::sendZeroTimerCallback, this) ;

        // Water timer
        timer_water = n.createTimer(ros::Duration(3), &ACTUATORS_ROS::sendWaterRequestTimerCallback, this) ;

        unsigned int i = 0 ;
    }

    void getConfig()
    {
        bool valid_config = true ;

        // Get params
        if (!ros::param::getCached("actuators_485/sp_path", device_config.sp_path))
        {
            ROS_FATAL_STREAM(_node_name << ": invalid parameters for actuators_485/sp_path in param server!") ;
            valid_config = false ;
        }
        if (!ros::param::getCached("actuators_485/sp_baud_rate", device_config.sp_baud_rate))
        {
            ROS_FATAL_STREAM(_node_name << ": invalid parameters for actuators_485/sp_baud_rate in param server!") ;
            valid_config = false ;
        }
        if (!ros::param::getCached("actuators_485/sp_char_size", device_config.sp_char_size))
        {
            ROS_FATAL_STREAM(_node_name << ": invalid parameters for actuators_485/sp_char_size in param server!") ;
            valid_config = false ;
        }
        if (!ros::param::getCached("actuators_485/sp_stop_bits", device_config.sp_stop_bits))
        {
            ROS_FATAL_STREAM(_node_name << ": invalid parameters for actuators_485/sp_stop_bits in param server!") ;
            valid_config = false ;
        }
        if (!ros::param::getCached("actuators_485/sp_parity", device_config.sp_parity))
        {
            ROS_FATAL_STREAM(_node_name << ": invalid parameters for actuators_485/sp_parity in param server!") ;
            valid_config = false ;
        }
        if (!ros::param::getCached("actuators_485/sp_flow_control", device_config.sp_flow_control))
        {
            ROS_FATAL_STREAM(_node_name << ": invalid parameters for actuators_485/sp_flow_control in param server!") ;
            valid_config = false ;
        }
        if (!ros::param::getCached("thrusters/request_info", device_config.request_info))
        {
            ROS_FATAL_STREAM(_node_name << ": invalid parameters for thrusters/request_info in param server!") ;
            valid_config = false ;
        }
        if (!ros::param::getCached("thrusters/max_step", device_config.max_step))
        {
            ROS_FATAL_STREAM(_node_name << ": invalid parameters for thrusters/max_step in param server!") ;
            valid_config = false ;
        }
        if (!ros::param::getCached("thrusters/symmetric", device_config.symmetric))
        {
            ROS_FATAL_STREAM(_node_name << ": invalid parameters for thrusters/symmetric in param server!") ;
            valid_config = false ;
        }

        // Addresses list
        try
        {
            XmlRpc::XmlRpcValue my_list_a ;
            if(ros::param::getCached("thrusters/addresses", my_list_a))
            {
                ROS_ASSERT(my_list_a.getType() == XmlRpc::XmlRpcValue::TypeArray) ;

                for (int32_t i = 0; i < my_list_a.size(); ++i)
                {
                    ROS_ASSERT(my_list_a[i].getType() == XmlRpc::XmlRpcValue::TypeInt) ;
                    device_config.addresses.push_back(static_cast<unsigned char>(static_cast<int>(my_list_a[i]))) ;
                }
            }
            else
            {
                ROS_FATAL_STREAM(_node_name << ": invalid parameters for thrusters/addresses in param server!") ;
                valid_config = false ;
            }
        }
        catch (std::exception& e)
        {
            ROS_FATAL_STREAM(_node_name << ": exception when loading thrusters/addresses: " << e.what()) ;
            valid_config = false ;
        }

        // Signs list
        try
        {
            XmlRpc::XmlRpcValue my_list_b ;
            if(ros::param::getCached("thrusters/signs", my_list_b))
            {
                ROS_ASSERT(my_list_b.getType() == XmlRpc::XmlRpcValue::TypeArray) ;

                for (int32_t i = 0; i < my_list_b.size(); ++i)
                {
                    ROS_ASSERT(my_list_b[i].getType() == XmlRpc::XmlRpcValue::TypeInt) ;
                    device_config.signs.push_back(static_cast<double>(static_cast<int>(my_list_b[i]))) ;
                }
            }
            else
            {
                ROS_FATAL_STREAM(_node_name << ": invalid parameters for thrusters/signs in param server!") ;
                valid_config = false ;
            }
        }
        catch (std::exception& e)
        {
            ROS_FATAL_STREAM(_node_name << ": exception when loading thrusters/signs: " << e.what()) ;
            valid_config = false ;
        }

        // Limiter list
        try
        {
            XmlRpc::XmlRpcValue my_list_c ;
            if(ros::param::getCached("thrusters/limiter", my_list_c))
            {
                ROS_ASSERT(my_list_c.getType() == XmlRpc::XmlRpcValue::TypeArray) ;

                for (int32_t i = 0; i < my_list_c.size(); ++i)
                {
                    ROS_ASSERT(my_list_c[i].getType() == XmlRpc::XmlRpcValue::TypeDouble) ;
                    device_config.limiter.push_back(static_cast<double>(my_list_c[i])) ;
                }
            }
            else
            {
                ROS_FATAL_STREAM(_node_name << ": invalid parameters for thrusters/limiter in param server!") ;
                valid_config = false ;
            }
        }
        catch (std::exception& e)
        {
            ROS_FATAL_STREAM(_node_name << ": exception when loading thrusters/limiter: " << e.what()) ;
            valid_config = false ;
        }

        // Fins address
        int fins_address ;
        if (!ros::param::getCached("fins/address", fins_address))
        {
            ROS_FATAL_STREAM(_node_name << ": invalid parameters for fins/address in param server!") ;
            valid_config = false ;
        }
        else
        {
            _fins_address = (unsigned char)fins_address ;
        }

        // Shutdown if not valid
        if (!valid_config)
        {
            ROS_FATAL_STREAM(_node_name << ": shutdown due to invalid config parameters!") ;
            ros::shutdown() ;
        }
    }

    bool openSerialPort()
    {
        if (!serial_port.isOpen())
        {
            try
            {
                serial_port.open(device_config.sp_path) ;
                serial_port.setBaudRate(cola2::io::SerialPort::baudRateFromInteger(device_config.sp_baud_rate)) ;
                serial_port.setCharSize(cola2::io::SerialPort::charSizeFromInteger(device_config.sp_char_size)) ;
                serial_port.setNumOfStopBits(cola2::io::SerialPort::numOfStopBitsFromInteger(device_config.sp_stop_bits)) ;
                serial_port.setParity(cola2::io::SerialPort::parityFromString(device_config.sp_parity)) ;
                serial_port.setFlowControl(cola2::io::SerialPort::flowControlFromString(device_config.sp_flow_control)) ;
            }
            catch (std::exception& e)
            {
                ROS_ERROR_STREAM(_node_name << ": error setting up the serial port: " << e.what()) ;
            }
        }
        else
        {
            ROS_ERROR_STREAM(_node_name << ": serial port already open!") ;
        }

        return serial_port.isOpen() ;
    }

    // Last setpoint executable for the main
    void finsSetToZero()
    {
        // Using the enable command to set fins to zero
        _fins_ptr->enableFins(_fins_address) ;
    }

    // Disable fins executable for the main
    void disableFinsAux()
    {
        _fins_ptr->disableFins(_fins_address) ;
    }

    void finsSetpointCallback( const cola2_control::Setpoints& fins_msg)
    {
        std::vector<double> fins_setpoint = fins_msg.setpoints ;

        // Apply a symmetric ramp to each fin
        int i ;
        for (i = 0 ; i < fins_setpoint.size() ; i++)
        {
            if (fins_setpoint[i] > _old_fins_setpoints[i] + 8)
            {
                fins_setpoint[i] = _old_fins_setpoints[i] + 8 ;
            }
            else if (fins_setpoint[i] < _old_fins_setpoints[i] - 8)
            {
                fins_setpoint[i] = _old_fins_setpoints[i] - 8 ;
            }

            _old_fins_setpoints[i] = fins_setpoint[i] ;
        }

        // Compute fins setpoints
        _fins_ptr->computeFinsSetpoint(_fins_address, ( fins_setpoint[0] + 60 ) * 2, ( fins_setpoint[1] + 60 ) * 2 ) ;
    }

    void sendWaterRequestTimerCallback(const ros::TimerEvent& event)
    {
            if (not _water)
            {
                _water = _fins_ptr->waterRequest(_fins_address) ;

                if (_water)
                {
                    ROS_FATAL_STREAM(_node_name << ": Water inside fins " << _water_offset) ;

                    if (_water_offset < 5)
                    {
                        _water_offset++ ;
                        _water = false ;
                    }
                    else
                    {
                        _water = true ;
                    }
                }
                else
                {
                    _water_offset = 0 ;
                }
            }

            if (_water)
            {
                diagnostic.add("water_inside_fins", "True") ;
                diagnostic.setLevel(diagnostic_msgs::DiagnosticStatus::ERROR, "Water inside fins") ;
                ROS_FATAL_STREAM(_node_name << ": WATER IN FINS COMPARTMENT! WATER IN FINS COMPARTMENT!") ;
            }
            else
            {
                diagnostic.add("water_inside_fins", "False") ;
                diagnostic.setLevel(diagnostic_msgs::DiagnosticStatus::OK, "No water") ;
            }
    }

    void thrustersSetpointCallback(const cola2_control::Setpoints& thrusters_msg)
    {
        _thrusters_ptr->thrusterSetpointCallback(thrusters_msg.setpoints, device_config.addresses, device_config.signs, device_config.limiter, device_config.symmetric, device_config.max_step, device_config.request_info) ;

        if (device_config.request_info)
        {
            thrusters_info.thruster_voltages = _thrusters_ptr->thrusters_info.thruster_voltages ;
            thrusters_info.thruster_currents = _thrusters_ptr->thrusters_info.thruster_currents ;
            thrusters_info.thruster_rpms = _thrusters_ptr->thrusters_info.thruster_rpms ;
            thrusters_info.thruster_controller_temperatures = _thrusters_ptr->thrusters_info.thruster_controller_temperatures ;
            thrusters_info.thruster_output_powers = _thrusters_ptr->thrusters_info.thruster_output_powers ;
            thrusters_info.thruster_warnings = _thrusters_ptr->thrusters_info.thruster_warnings ;
            thrusters_info.thruster_errors = _thrusters_ptr->thrusters_info.thruster_errors ;
            thrusters_info.total_ms_in_driver = _thrusters_ptr->thrusters_info.total_ms_in_driver ;
            thrusters_info.setpoints = _thrusters_ptr->thrusters_info.setpoints ;
            thrusters_info.processed_setpoints = _thrusters_ptr->thrusters_info.processed_setpoints ;

            thruster_info_pub.publish(thrusters_info) ;
        }
    }

    void sendZeroTimerCallback(const ros::TimerEvent& event)
    {
        if (boost::posix_time::time_period(lastSetpointCallback, boost::posix_time::microsec_clock::universal_time()).length().total_microseconds() > 200000)
        {
            _thrusters_ptr->sendZeros(device_config.addresses) ;
        }
        lastSetpointCallback = boost::posix_time::microsec_clock::universal_time() ;
    }

    // Send zeros executable for the main
    void sendZeroAux()
    {
        _thrusters_ptr->sendZeros(device_config.addresses) ;
    }
} ;

sig_atomic_t volatile running = 1 ;
void stopHandler(int sig)
{
    running = 0 ;
}

int main(int argc, char **argv)
{
    ros::init(argc, argv, "actuators_485", ros::init_options::NoSigintHandler) ;
    signal(SIGINT, stopHandler) ;

    // Node name
    std::string node_name ;
    node_name = ros::this_node::getName() ;

    // Init. FINS_485 object
    FINS_485 *fins_ptr ;
    fins_ptr = new FINS_485(node_name) ;

    // Init. THRUSTERS_485 object
    THRUSTERS_485 *thrusters_ptr ;
    thrusters_ptr = new THRUSTERS_485(node_name) ;

    // Init. ACTUATORS_485 object
    ACTUATORS_ROS actuators(node_name, fins_ptr, thrusters_ptr) ;

    // Get parameters, enable fins and init thrusters
    actuators.init() ;

    // Show message
    ROS_INFO_STREAM(node_name << " :initialized") ;

    // Iterate while not shutdown requested
    while (running)
    {
        ros::spinOnce() ;
        ros::Duration(0.01).sleep() ;
    }

    // Pose fins to zero
    actuators.finsSetToZero() ;

    // Let the fins go to zero
    ros::Duration(0.01).sleep() ;

    // Disable fins
    actuators.disableFinsAux() ;

    // Send zeros to thrusters
    actuators.sendZeroAux() ;

    // Shutdown
    ros::shutdown() ;
    return 0 ;
}
