//////////////////////////////////////////////////////////////////////////////////////////////////////////
////                                          fins_program.C                                          ////
////                                           version: 1.5                                           ////
////                                          date: 07/05/15                                          ////
////                                        author: ric Pairet                                       ////
////                                                                                                  ////
////  This program is able to control up to 2 servomotors and detect the existence of not desired     ////
////  water in the fins servomotors compartment                                                       ////
////                                                                                                  ////
////  - The recognised ASCII commands for this program are:                                           ////
////                                                                                                  ////
////  v\r            to ask for the firmware version                                                  ////
////                                                                                                  ////
////  MinLXXXX\r     to configure the equivalent time to the minimum angle for the left fin           ////
////  ZerLXXXX\r     to configure the equivalent time to the horizontal position for the left fin     ////                   
////  MaxLXXXX\r     to configure the equivalent time to the maximum angle for the left fin           ////
////  RanLXXXX\r     to configure the left fin range in degrees                                       ////
////  MinRXXXX\r     to configure the equivalent time to the minimum angle for the right fin          ////
////  ZerRXXXX\r     to configure the equivalent time to the horizontal position for the right fin    ////                      
////  MaxRXXXX\r     to configure the equivalent time to the maximum angle for the right fin          ////
////  RanRXXXX\r     to configure the right fin range in degrees                                      ////
////                                                                                                  ////
////  Where XXXX is a maximum 4 digits number which contents the value of the configured parameter    ////
////                                                                                                  ////
////  - The recognised binary commands for this program are:                                          ////
////                                                                                                  ////
////  253 + ADDRESS + LF + RF + BCC    to control the fins position                                   ////
////                                                                                                  ////
////  Where:                                                                                          ////
////     ADDRESS is the specified address in the 'fins_program.h' file                                ////
////     LF is two times the desired angle in degrees for the left fin                                ////
////     RF is two times the desired angle in degrees for the right fin                               ////
////     BCC = ADDRESS xor LF xor RF                                                                  ////
////                                                                                                  ////
////  254 + ADDRESS + 0 + 0      to disable the fins control                                          ////
////  254 + ADDRESS + 1 + 1      to enable the fins control                                           ////
////  254 + ADDRESS + 2 + 2      to ask for water existence                                           ////
////  254 + ADDRESS + 10 + 10    to enable the ASCII commands                                         ////
////                                                                                                  ////
////  Where:                                                                                          ////
////     ADDRESS is the specified address in the 'fins_program.h' file                                ////
//////////////////////////////////////////////////////////////////////////////////////////////////////////

//Attach of required libraries
#include <fins_program.h>
#include <inputRaw.h>
#include <stdlib.h>

//Global variables declaration
unsigned int8 delta_time;
unsigned int8 servomotor = 0;
unsigned int8 times_array[2];

float LF_input_angle;
float LF_max;
float LF_min;
float LF_range;
float LF_time_angle;
float LF_zero;
float RF_input_angle;
float RF_max;
float RF_min;
float RF_range;
float RF_time_angle;
float RF_zero;

//TIMER0 interruption -- outputs to low
#int_TIMER0
void  TIMER0_isr(void) 
{  
   disable_interrupts(INT_TIMER0);
   clear_interrupt(INT_TIMER0);
   
   if ( servomotor == 1 )
   {
      output_low(SERVO_L);
      set_timer0(delta_time);
      
      enable_interrupts(INT_TIMER0);
   }
   else if ( servomotor == 2 )
   {
      output_low(SERVO_R);
      set_timer0(delta_time);
      
      enable_interrupts(INT_TIMER0);
   }
   else if ( servomotor == 0 )
   {
      output_low(SERVO_L);
      output_low(SERVO_R);
   }
   servomotor = 0;
}

//TIMER2 interruption -- outputs to high
#int_TIMER2
void  TIMER2_isr(void) 
{
   clear_interrupt(INT_TIMER0);
 
   if ( times_array[0] > ( times_array[1] + 1 ) )
   {
      set_timer0(times_array[0]);
      
      delta_time = 256 - ( times_array[0] - times_array[1] ) + 1;
      
      servomotor = 1;
   }
   else if ( ( times_array[0] + 1 ) < times_array[1] )
   {
      set_timer0(times_array[1]);
      
      delta_time = 256 - ( times_array[1] - times_array[0] ) + 1;
      
      servomotor = 2;
   }
   else
   {
      set_timer0(times_array[0]);
      
      servomotor = 0;
   }
   
   output_high(SERVO_L);
   output_high(SERVO_R);
   
   enable_interrupts(INT_TIMER0);   
}

//Serial port interruption
#int_RDA
void  RDA_isr(void) 
{
   putInputRaw(getc());
}

//Function to compute the setpoints
void computeSetpoints()
{  
   //Calculating the equivalent time to the desired angle for the left fin
   if ( LF_input_angle <= ( LF_range / 2 ) )
   {
      LF_time_angle = ( ( ( ( LF_input_angle * ( LF_zero - LF_min ) ) / LF_range ) * 2 ) + LF_min );
      
      //Avoiding invalid time values
      if ( LF_time_angle < LF_min )
      {
         LF_time_angle = LF_min;
      }
   }
   else // LF_input_angle > ( LF_range / 2 )
   {
      LF_input_angle -= ( LF_range / 2 );
      LF_time_angle = ( ( ( ( LF_input_angle * ( LF_max - LF_zero ) ) / LF_range ) * 2 ) + LF_zero );
      
      //Avoiding invalid time values
      if ( LF_time_angle > LF_max )
      {
         LF_time_angle = LF_max;
      }
   }
   
   //Horizontal mirror of the desired angle for the right fin according to the mechanical disposition
   if ( RF_input_angle < RF_range )
   {
      RF_input_angle = RF_range - RF_input_angle;
   }
   else 
   {
      RF_input_angle = 0;
   }
   
   //Calculating the equivalent time to the desired angle for the right fin
   if ( RF_input_angle <= ( RF_range / 2 ) )
   {
      RF_time_angle = ( ( ( ( RF_input_angle * ( RF_zero - RF_min ) ) / RF_range ) * 2 ) + RF_min );
      
      //Avoiding invalid time values
      if ( RF_time_angle < RF_min )
      {
         RF_time_angle = RF_min;
      }
   }
   
   else // RF_input_angle > ( RF_range / 2 )
   {
      RF_input_angle -= ( RF_range / 2 );
      RF_time_angle = ( ( ( ( RF_input_angle * ( RF_max - RF_zero ) ) / RF_range ) * 2 ) + RF_zero );
      
      //Avoiding invalid time values
      if ( RF_time_angle > RF_max )
      {
         RF_time_angle = RF_max;
      }
   }
}

//Main function
void main()
{
   //PIC initialising
   setup_timer_0(RTCC_INTERNAL|RTCC_DIV_32|RTCC_8_bit);     //2.96 ms overflow, 11.5625us resolution
   setup_timer_2(T2_DIV_BY_4,253,9);                        //357 us overflow, 3.3 ms interrupt  
   setup_timer_4(T4_DISABLED,0,1);
   setup_timer_6(T6_DISABLED,0,1);
   setup_comparator(NC_NC_NC_NC);
   disable_interrupts(INT_TIMER0);
   disable_interrupts(INT_TIMER2);
   disable_interrupts(INT_TIMER4);
   disable_interrupts(INT_TIMER6);
   enable_interrupts(INT_RDA);
   enable_interrupts(GLOBAL);
   
   //Local variables declaration
   int1 enable_ascii = 0;
   int1 water_detected = 0;
   
   unsigned int8 alpha_counter = 0;
   unsigned int8 bcc;
   unsigned int8 configure_paramater[4];
   unsigned int8 high_part;
   unsigned int8 low_part;
   unsigned int8 new_data;
   unsigned int8 number_counter = 0;
   unsigned int8 request;
   unsigned int8 selected_paramater;
   unsigned int8 sent_command;
   unsigned int8 state = COMMAND;
   
   unsigned int16 times_water_detected = 0;
   unsigned int16 value;
  
   char paramater_value[4];
   
   //External functions call
   inputRawInit();
   RCSTA = 0x90;
      
   //Recuperation of essential parameters from EEPROM
   LF_min = make16(read_eeprom(0),read_eeprom(1));
   LF_zero = make16(read_eeprom(2),read_eeprom(3));
   LF_max = make16(read_eeprom(4),read_eeprom(5));
   LF_range = make16(read_eeprom(6),read_eeprom(7));
   RF_min = make16(read_eeprom(8),read_eeprom(9));   
   RF_zero = make16(read_eeprom(10),read_eeprom(11));
   RF_max = make16(read_eeprom(12),read_eeprom(13));
   RF_range = make16(read_eeprom(14),read_eeprom(15));
      
   //Configuration of serial port
   output_bit(DIRECTION, RECEIVE);
   
   //Main loop
   while(TRUE)
   {  
      //Check water sensor
      if ( times_water_detected < 34695 )
      {
         if ( input(WATER_SENSOR) == 0 )
         {
            times_water_detected += 1;
         }
         else
         {
            times_water_detected = 0;
         }
      }
      
      //Check serial port
      if (RCSTA & 0x02) // OVERRUN ERROR
      {
         RCSTA = 0x80;  // CLEAR CREN BIT
         RCSTA = 0x90;
      }
      
      //PC data protocol
      if ( !inputRawEmpty() )
      {
         new_data = getFromInputRaw();
         
         switch ( state )
         {
            case COMMAND:
               if ( ( ( new_data == 253 ) || ( new_data == 254 ) ) && ( alpha_counter == 0 ) ) 
               {
                  bcc = 0;
                  sent_command = new_data;
                  state = ADDRESS;
               }
               else if ( enable_ascii )
               {
                  if ( ( new_data == 'v' ) && ( alpha_counter == 0 ) )
                  {
                     state = CONFIRMATION;
                     break;
                  }
                  else if ( !( isalpha(new_data) ) && ( alpha_counter >= 1 ) )
                  {
                     //ERROR
                     alpha_counter = 0;
                     state = COMMAND;
                     break;
                  }
                  else if ( isalpha(new_data) )
                  {
                     configure_paramater[alpha_counter] = new_data;
                     alpha_counter++;
                     
                     if ( alpha_counter == 4 )
                     {                  
                        if ( ( configure_paramater[0] == 'M' ) && ( configure_paramater[1] == 'i' ) && ( configure_paramater[2] == 'n' ) && ( configure_paramater[3] == 'L' ) )
                        {
                           selected_paramater = 0;
                        }
                        else if ( ( configure_paramater[0] == 'Z' ) && ( configure_paramater[1] == 'e' ) && ( configure_paramater[2] == 'r' ) && ( configure_paramater[3] == 'L' ) )
                        {
                           selected_paramater = 2;
                        }
                        else if ( ( configure_paramater[0] == 'M' ) && ( configure_paramater[1] == 'a' ) && ( configure_paramater[2] == 'x' ) && ( configure_paramater[3] == 'L' ) )
                        {
                           selected_paramater = 4;
                        }
                        else if ( ( configure_paramater[0] == 'R' ) && ( configure_paramater[1] == 'a' ) && ( configure_paramater[2] == 'n' ) && ( configure_paramater[3] == 'L' ) )
                        {
                           selected_paramater = 6;
                        }
                        else if ( ( configure_paramater[0] == 'M' ) && ( configure_paramater[1] == 'i' ) && ( configure_paramater[2] == 'n' ) && ( configure_paramater[3] == 'R' ) )
                        {
                           selected_paramater = 8;
                        }
                        else if ( ( configure_paramater[0] == 'Z' ) && ( configure_paramater[1] == 'e' ) && ( configure_paramater[2] == 'r' ) && ( configure_paramater[3] == 'R' ) )
                        {
                           selected_paramater = 10;
                        }
                        else if ( ( configure_paramater[0] == 'M' ) && ( configure_paramater[1] == 'a' ) && ( configure_paramater[2] == 'x' ) && ( configure_paramater[3] == 'R' ) )
                        {
                           selected_paramater = 12;
                        }
                        else if ( ( configure_paramater[0] == 'R' ) && ( configure_paramater[1] == 'a' ) && ( configure_paramater[2] == 'n' ) && ( configure_paramater[3] == 'R' ) )
                        {
                           selected_paramater = 14;
                        }
                        else
                        {
                           //ERROR
                           alpha_counter = 0;
                           state = COMMAND;
                           break;
                        }
                        alpha_counter = 0;
                        number_counter = 0;
                        state = CONFIGURATION;
                     }
                  }
                  else
                  {
                     state = COMMAND;
                  }
               }
               else
               {
                  state = COMMAND;
               }
            break;
            
            case ADDRESS:
               if ( new_data == ADDRESS_ID )
               {
                  bcc = ADDRESS_ID;
                  state = PARAMETER1;
               }
               else
               {
                  state = COMMAND;
               }
            break;
            
            case PARAMETER1:               
               if ( sent_command == 253 )
               {
                  LF_input_angle = ( new_data / 2 );    // Double resolution
                  bcc ^= new_data;
                  state = PARAMETER2;
               }
               else if ( sent_command == 254 )
               {
                  request = new_data;
                  state = PARAMETER2;
               }
               else
               {
                  state = COMMAND;
               }
            break;
            
            case PARAMETER2:    
               if ( sent_command == 253 )
               {
                  RF_input_angle = ( new_data / 2 );   // Double resolution
                  bcc ^= new_data;
                  state = CONTROL_BIT; 
               }
               else if ( ( sent_command == 254 ) && ( request == new_data ) )
               {                 
                  switch ( request )
                  {
                     case 0:                       
                        disable_interrupts(INT_TIMER0);
                        disable_interrupts(INT_TIMER2);
                     break;
                     
                     case 1:
                        //Arranging the table with the desired position
                        times_array[0] = (int8)( 256 - (int16)( LF_zero / 11.5625 ) );
                        times_array[1] = (int8)( 256 - (int16)( RF_zero / 11.5625 ) );

                        enable_interrupts(INT_TIMER2);
                        
                     break;
                     
                     case 2:
                        if ( times_water_detected == 34695 ) // 34695 iteration is aprox 450 ms
                        {
                           water_detected = 1;
                           
                           // Once answered the request -> reset
                           times_water_detected = 0;
                        }
                        else
                        {
                           water_detected = 0;
                        }
                                          
                        output_bit(DIRECTION, TRANSMIT);                       
                        printf("%c",254);
                        printf("%c",ADDRESS_ID);
                        printf("%c",2);
                        printf("%c",water_detected);              
                        printf("%c",water_detected);
                        printf("%c",ADDRESS_ID ^ 2);
                        delay_us(348);    
                        output_bit(DIRECTION, RECEIVE);
                     break;
                     
                     case 10:
                        if ( enable_ascii == 1 )
                        {
                           enable_ascii = 0;
                           
                           output_bit(DIRECTION, TRANSMIT);
                           printf("Config. disabled\n");
                           delay_us(348);                                    
                           output_bit(DIRECTION, RECEIVE);
                        }
                        else if ( enable_ascii == 0 )
                        {
                           enable_ascii = 1;
                           
                           output_bit(DIRECTION, TRANSMIT);
                           printf("Config. enabled\n");
                           delay_us(348);                                    
                           output_bit(DIRECTION, RECEIVE);
                        }
                     break;
                     
                     default:
                        state = COMMAND;
                     break;
                  }
                  state = COMMAND;
               }
               else
               {
                  state = COMMAND;
               }
            break;
            
            case CONTROL_BIT:
               if ( ( bcc == new_data ) && ( sent_command == 253 ) )
               {  
                  //
                  computeSetpoints();
                                                                                         
                  //Arranging the table with the desired position
                  times_array[0] = (int8)( 256 - (int16)( LF_time_angle / 11.5625 ) );
                  times_array[1] = (int8)( 256 - (int16)( RF_time_angle / 11.5625 ) );
               }
               state = COMMAND;
            break;
            
            case CONFIRMATION:
               if ( new_data == '\r' )
               {  
                  output_bit(DIRECTION, TRANSMIT);
                  printf("%s\n",FIRMWARE_VERSION);
                  delay_us(348);
                  output_bit(DIRECTION, RECEIVE);
                  state = COMMAND;
               }
               else
               {
                  //ERROR
                  output_bit(DIRECTION, TRANSMIT);
                  printf("E\n");
                  delay_us(348);
                  output_bit(DIRECTION, RECEIVE);
                  state = COMMAND;
               }
            break;
            
            case CONFIGURATION:
               if ( new_data != '\r' )
               {
                  if ( isdigit(new_data) && (number_counter < 4) )
                  {
                     paramater_value[number_counter] = new_data;
                     number_counter++;
                  }
                  else
                  {
                     //ERROR
                     output_bit(DIRECTION, TRANSMIT);
                     printf("E\n");
                     delay_us(348);
                     output_bit(DIRECTION, RECEIVE);
                     number_counter = 0;
                     state = COMMAND;
                  }
                  if ( ( state == CONFIGURATION ) && ( number_counter < 4 ) )
                  {
                     paramater_value[number_counter+1] = '\0';
                  }
               }
               else
               {
                  if ( number_counter != 0 )
                  {  
                     value = atol(paramater_value);
                     high_part = make8(value,1);
                     low_part = make8(value,0);
                     
                     switch ( selected_paramater )
                     {
                        case MINL:
                           LF_min = value;
                        break;
                        
                        case ZERL:
                           LF_zero = value;
                        break;
                        
                        case MAXL:
                           LF_max = value;
                        break;
                        
                        case RANL:
                           LF_range = value;
                        break;
                        
                        case MINR:
                           RF_min = value;
                        break;
                        
                        case ZERR:
                           RF_zero = value;
                        break;
                        
                        case MAXR:
                           RF_max = value;
                        break;
                        
                        case RANR:
                           RF_range = value;
                        break;
                     }
                     write_eeprom(selected_paramater,high_part);
                     write_eeprom(selected_paramater+1,low_part);
                     
                     output_bit(DIRECTION, TRANSMIT);
                     printf("OK\n");
                     delay_us(348);
                     output_bit(DIRECTION, RECEIVE);
                     
                     state = COMMAND;
                  }
                  else
                  {  
                     //ERROR
                     output_bit(DIRECTION, TRANSMIT);
                     printf("E\n");
                     delay_us(348);
                     output_bit(DIRECTION, RECEIVE);
                     state = COMMAND;
                  }
                  number_counter = 0;
               }
            break;
                        
            default:
               state = COMMAND;
            break;
         }                           
      }                    
   }
}
